"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This package implements agents for the docking and inspection
environments.

Agents are actors which control [platforms](../platforms/index.md)
within an environment. They make decisions based on a learned or
programmed policy given their current observation space. The agent's
actions influence the environment via platform controls.
"""
