"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This package implements initializers for the safe autonomy simulators.

Initializers are responsible for setting up the initial conditions
of platforms in a simulation. Given a set of sampled parameters,
initializers return a set of initialization parameters expected by
the platforms and their associated simulation entities.
"""
