"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

The `safe-autonomy-sims` package provides reinforement learning environments
built using the CoRL framework.

The included environments define safety-critical spacecraft docking and
spacecraft inspection tasks. The intent of these environments is to
provide a baseline for training RL agents to safely perform these tasks.
"""
